/**
 * Fractal ADL Parser
 * Copyright (C) 2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.adl.error;

import org.objectweb.fractal.adl.Node;

/**
 * An error locator for locating error in an AST.
 */
public class NodeErrorLocator implements ErrorLocator {

  private final Node   node;
  private final int    line;
  private final int    column;
  private final String inputFilePath;

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Creates a {@link NodeErrorLocator} for the given node.
   * 
   * @param node an AST node.
   */
  public NodeErrorLocator(final Node node) {
    if (node == null) throw new IllegalArgumentException("node can't be null");
    this.node = node;
    final String source = node.astGetSource();
    if (source == null) {
      inputFilePath = null;
      line = -1;
      column = -1;
    } else {
      final int i = source.lastIndexOf(':');
      if (i == -1) {
        inputFilePath = source;
        line = -1;
        column = -1;
      } else {
        inputFilePath = source.substring(0, i);
        final int j = source.indexOf('-', i);
        if (j == -1) {
          int lineNumer;
          try {
            lineNumer = Integer.parseInt(source.substring(i + 1));
          } catch (final NumberFormatException e) {
            lineNumer = -1;
          }
          line = lineNumer;
          column = -1;
        } else {
          int lineNumber;
          try {
            lineNumber = Integer.parseInt(source.substring(i + 1, j));
          } catch (final NumberFormatException e) {
            lineNumber = -1;
          }
          line = lineNumber;
          int columnNumber;
          try {
            columnNumber = Integer.parseInt(source.substring(j + 1));
          } catch (final NumberFormatException e) {
            columnNumber = -1;
          }
          column = columnNumber;
        }
      }
    }
  }

  /**
   * Returns the node located by this locator.
   * 
   * @return the node located by this locator.
   */
  public Node getNode() {
    return node;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ErrorLocator interface
  // ---------------------------------------------------------------------------

  public int getBeginColumn() {
    return column;
  }

  public int getBeginLine() {
    return line;
  }

  public int getEndColumn() {
    return -1;
  }

  public int getEndLine() {
    return -1;
  }

  public String getInputFilePath() {
    return inputFilePath;
  }

  public String getLocation() {
    if (inputFilePath == null) return "unknown";
    String location = inputFilePath;
    if (line != -1) location += ":" + line;
    if (column != -1) location += "-" + column;
    return location;
  }

  // ---------------------------------------------------------------------------
  // Overridden methods
  // ---------------------------------------------------------------------------

  @Override
  public String toString() {
    return getLocation();
  }
}
